#version 120

#define FOG_ATTENUATION 0.035 // Bigger values -> more fog [0.01 0.015 0.02 0.025 0.03 0.035 0.04 0.045 0.05 0.055 0.06 0.065 0.07 0.075 0.08 0.085 0.09 0.095 0.1]
#define FOG_RED 0.84 // Amount of red in fog colour [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.84 0.9 0.95 1.0]
#define FOG_GREEN 0.84 // Amount of green in fog colour [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.84 0.9 0.95 1.0]
#define FOG_BLUE 0.84 // Amount of blue in fog colour [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.84 0.9 0.95 1.0]
#define LIGHT_SCATTERING_INTENSITY 0.5 //The blur that appears on distance objects to mimix real-life light scattering from fog [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.84 0.9 0.95 1.0]
#define BLOOM_INTENSITY 0.7 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]

#define SSR_QUALITY 2 //The accuracy of screen-space reflections [0 1 2 3]
#define MAX_REFLECTION_DIST 65 //Distance where reflections are no longer calculated [35 65 145]

#define UNDERWATER_REFRACTION
#define WATER_FOG_RED 0.0 // Amount of red in fog colour [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.85 0.9 0.95 1.0]
#define WATER_FOG_GREEN 0.05 // Amount of green in fog colour [0.0 0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.855 0.9 0.95 1.0]
#define WATER_FOG_BLUE 0.08 // Amount of blue in fog colour [0.0 0.08 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7 0.75 0.8 0.85 0.9 0.95 1.0]

#define GODRAYS
#define GODRAY_SAMPLES 12 //The quality of volumetric light scattering [12 24]

#define SSAO //Screen-space ambient occlusion. Replicates the way light has trouble reaching small cracks and spaces.
#define LUT 0//[0 1 2]
float bayer2(vec2 a){
    a = floor(a);
    return fract( dot(a, vec2(.5, a.y * .75)) );
}
#define bayer4(a)   (bayer2( .5*(a))*.25+bayer2(a))
#define bayer8(a)   (bayer4( .5*(a))*.25+bayer2(a))
#define bayer16(a)  (bayer8( .5*(a))*.25+bayer2(a))
#define bayer32(a)  (bayer16(.5*(a))*.25+bayer2(a))
#define bayer64(a)  (bayer32(.5*(a))*.25+bayer2(a))
#define bayer128(a) (bayer64(.5*(a))*.25+bayer2(a))
#define bayer256(a) (bayer128(.5*(a))*.25+bayer2(a))

varying vec4 texcoord;

uniform sampler2D gcolor;
uniform sampler2D colortex1;//Bloom'd things
uniform sampler2D colortex2;//normals
uniform sampler2D colortex3;//specularity 
uniform sampler2D colortex4;//hand colour 
uniform sampler2D noisetex;
uniform sampler2D gaux1;
uniform sampler2D gaux2;
uniform sampler2D gaux3;
uniform sampler2D gaux4;
uniform float viewWidth;
uniform float viewHeight;
uniform vec2 pixelSize;
uniform int worldTime;
uniform vec3 cameraPosition;
uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;
uniform mat4 gbufferProjection;
uniform float aspectRatio;
uniform float near;  
uniform float far;   
uniform int isEyeInWater;  
uniform float frameTimeCounter; 
uniform float rainStrength;  
uniform vec3 sunPosition;
uniform vec3 moonPosition;
uniform float wetness;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

const int QUALITY = 4;
const int DIRECTIONS = 8;
const float TWOPI = 6.28318530718;

const vec3 NIGHT_FOG_COLOR = vec3(0.0784);

const vec4 sunColor = vec4(1.0, 0.8, 0.70, 1.0);
const vec4 moonColor = vec4(0.5, 0.5, 1.0, 1.0);

const float waterIndex = 1.5;
const float waterReflCoeff = 0.28;//pow((1.0 - waterIndex)/(1.0 + waterIndex), 2);

const float wetnessHalflife = 20.0;
const float drynessHalflife = 100.0;

const float minWetness = 0.1;

const int noiseTextureResolution = 64;


/*
const int colortex2Format = RGBA16;
const int gaux2Format = RGBA16;
const int gaux3Format = RGBA16;
const int gaux4Format = RGBA16;
const int gcolorFormat = RGBA16;
*/

//   const float total_strength = 1.0;
//   const float base = 0.2;
  
//   const float area = 0.0075;
//   const float falloff = 0.000001;
  
//   const float radius = 0.0002;
  
//   const float AOSamples = 16.0;
//   const vec3 sample_sphere[16] = vec3[](
//       vec3( 0.5381, 0.1856,-0.4319), vec3( 0.1379, 0.2486, 0.4430),
//       vec3( 0.3371, 0.5679,-0.0057), vec3(-0.6999,-0.0451,-0.0019),
//       vec3( 0.0689,-0.1598,-0.8547), vec3( 0.0560, 0.0069,-0.1843),
//       vec3(-0.0146, 0.1402, 0.0762), vec3( 0.0100,-0.1924,-0.0344),
//       vec3(-0.3577,-0.5301,-0.4358), vec3(-0.3169, 0.1063, 0.0158),
//       vec3( 0.0103,-0.5869, 0.0046), vec3(-0.0897,-0.4940, 0.3287),
//       vec3( 0.7119,-0.0154,-0.0918), vec3(-0.0533, 0.0596,-0.5411),
//       vec3( 0.0352,-0.0631, 0.5460), vec3(-0.4776, 0.2847,-0.0271)
//   );

float ld(float depth) {
   return (2.0 * near) / (far + near - depth * (far - near));
}
  float dither5x3()
{
	const int ditherPattern[15] = int[15](
		 9, 3, 7,12, 0,
		11, 5, 1,14, 8,
		 2,13,10, 4, 6);

    vec2 position = floor(mod(vec2(texcoord.s * viewWidth,texcoord.t * viewHeight), vec2(5.0,3.0)));

	int dither = ditherPattern[int(position.x) + int(position.y) * 5];

	return float(dither) / 15.0f;
}
/* Im too tired to write my own ssao so credit to Capt Tatsu for this*/
float calcAO(vec2 coord){
	float ao = 0.0;
	
	const int aoloop = 3;
	const int aoside = 3;
	float radius = 1.1/aoloop;
	float dither = bayer64(coord * vec2(viewWidth, viewHeight));
	float dither2 = fract(dither5x3()-dither);
	float d = texture2D(depthtex1,coord).r;
	float hand = float(d < 0.56);
	d = ld(d);
	const float piangle = 0.0174603175;
	float rot = 180/aoside*dither2;
	float size = radius*dither;
	float sd = 0.0;
	float angle = 0.0;
	float dist = 0.0;
	vec2 scale = vec2(1.0/aspectRatio,1.0) * gbufferProjection[1][1] / (2.74747742 * max(far*d,6.0));
	
	for (int i = 0; i < aoloop; i++) {
		for (int j = 0; j < aoside; j++) {
			sd = ld(texture2D(depthtex1,coord+vec2(cos(rot*piangle),sin(rot*piangle)) * size * scale).r);
			float sample = far*(d-sd)/size;
			if(hand > 0.5) sample *= 1024.0;
			angle = clamp(0.5-sample,0.0,1.0);
			dist = clamp(0.0625*sample,0.0,1.0);

			sd = ld(texture2D(depthtex1,coord-vec2(cos(rot*piangle),sin(rot*piangle)) * size * scale).r);
			sample = far*(d-sd)/size;
			if(hand > 0.5) sample *= 1024.0;
			angle += clamp(0.5-sample,0.0,1.0);
			dist += clamp(0.0625*sample,0.0,1.0);
			
			ao += clamp(angle+dist,0.0,1.0);
			rot += 180.0/aoside;
		}
		rot += 180.0/aoside;
		size += radius;
		angle = 0.0;
		dist = 0.0;
	}
	ao /= aoloop*aoside;
	
	return clamp((ao + 0.1)*1.05, 0.0, 1.0);
}

// float calcAOa(vec2 coord) {

    
//      vec3 random = normalize( texture2D(noisetex, coord * 4.0).rgb );
  
//   float depth = texture2D(depthtex1, coord).r;
 
//   vec3 position = vec3(coord, depth);
//   //vec3 position = getPointInViewOpaque(coord);
//   vec3 normal = texture2D(colortex2, coord).rgb;
//   normal = normal * 2.0 - 1.0;
//   float radius_depth = radius/depth;
//   float occlusion = 0.0;
//   for(int i=0; i < 16; i++) {
  
//     vec3 ray = radius_depth * reflect(sample_sphere[i], random);
//     vec3 hemi_ray = position + sign(dot(ray,normal)) * ray;
    
//     float occ_depth = texture2D(depthtex1, clamp(hemi_ray.xy, 0.0, 1.0)).r;
//     float difference = depth - occ_depth;
    
//     occlusion += step(falloff, difference) * (1.0-smoothstep(falloff, area, difference));
//   }
  
//   float ao = total_strength * occlusion * (1.0 / AOSamples);
//   return clamp(ao + base, 0.0, 1.0);
// }

float map(float n, float start1, float stop1, float start2, float stop2) {
    return ((n - start1) / (stop1 - start1)) * (stop2 - start2) + start2;
}

vec3 getPointInView(vec2 screenPos) {
    vec4 pos;
    pos.xy = screenPos;
    pos.z = texture2D(depthtex0, screenPos).r;
    pos.w = 1.0;
    pos.xyz = pos.xyz * 2.0 - 1.0; //convert from the 0-1 range to the -1 to +1 range
    pos = gbufferProjectionInverse * pos;
    pos.xyz /= pos.w;

    return pos.xyz;
}

vec3 getPointInViewOpaque(vec2 screenPos) {
    vec4 pos;
    pos.xy = screenPos;
    pos.z = texture2D(depthtex1, screenPos).r;
    pos.w = 1.0;
    pos.xyz = pos.xyz * 2.0 - 1.0; //convert from the 0-1 range to the -1 to +1 range
    pos = gbufferProjectionInverse * pos;
    pos.xyz /= pos.w;
    return pos.xyz;
}

float getSceneDist(vec2 screenPos) {


    return length(getPointInViewOpaque(screenPos));
}

float getFogAmount(float d) {
    return exp(-d*FOG_ATTENUATION);
}

vec3 getFogColor() {
    float fogType;//0 = night, 1 = day

    if(worldTime > 22000 && worldTime < 23999) {
        //From night to day
        fogType = map(float(worldTime), 22000.0, 23999.0, 0.0, 1.0);
    } else if(worldTime > 12000 && worldTime < 14000) {
        //From day to night
        fogType = map(float(worldTime), 12000.0, 14000.0, 1.0, 0.0);
    } else if(worldTime >= 0 && worldTime <= 12000) {
        //Day
        fogType = 1.0;
    } else {
        fogType = 0.0;
    }

    vec3 DAY_FOG_COLOR = vec3(FOG_RED, FOG_GREEN, FOG_BLUE);

    vec3 col = mix(NIGHT_FOG_COLOR, DAY_FOG_COLOR, fogType);
    return col;
}

vec3 doFog(vec3 color, vec2 pos) {
    float dist = getSceneDist(pos);

    float f;
    if(dist > 0.0) {
        f = getFogAmount(dist);
    } else {
        f = 1.0;
    }
    //Apply Fog
    vec3 fogColor;
    vec3 WATER_FOG_COLOR = vec3(WATER_FOG_RED, WATER_FOG_GREEN, WATER_FOG_BLUE);

    if(isEyeInWater > 0) {
        if(texture2D(colortex3, pos).r > 0.999) {
            fogColor = mix(WATER_FOG_COLOR, getFogColor(), 0.1);
            color *= 0.5;
        } else {
            fogColor = WATER_FOG_COLOR;
            f = exp(-dist*0.055);
        }
    } else {
        fogColor = getFogColor();
    }


    return mix(color, fogColor, 1.0 - f);
}

/* Gaussian blur code from RRe36, Kappa shader*/
const float gauss9w[9] = float[9] (
     0.0779, 0.12325, 0.0779,
    0.12325, 0.1954,  0.12225,
     0.0779, 0.12325, 0.0779
);

const vec2 gauss9o[9] = vec2[9] (
    vec2(1.0, 1.0), vec2(0.0, 1.0), vec2(-1.0, 1.0),
    vec2(1.0, 0.0), vec2(0.0, 0.0), vec2(-1.0, 0.0),
    vec2(1.0, -1.0), vec2(0.0, -1.0), vec2(-1.0, -1.0)
);

vec3 gauss9(sampler2D tex, vec2 coord, float sigma) {
    vec2 pixelRad = pixelSize*sigma*4.0;
    vec4 col = vec4(0.0);

    for (int i = 0; i<9; i++) {
        vec2 bcoord = coord + gauss9o[i]*pixelRad;
        col += texture2D(tex, bcoord)*gauss9w[i];
    }
    return col.rgb;
}
vec3 blur9(sampler2D image, vec2 coord, vec2 direction) {
    vec2 uv = coord;
    vec2 resolution = vec2(viewWidth, viewHeight);
    vec3 color = vec3(0.0);
    vec2 off1 = vec2(1.3846153846) * direction;
    vec2 off2 = vec2(3.2307692308) * direction;
    color += texture2D(image, uv).rgb * 0.2270270270;
    color += texture2D(image, uv + (off1 / resolution)).rgb * 0.3162162162;
    color += texture2D(image, uv - (off1 / resolution)).rgb * 0.3162162162;
    color += texture2D(image, uv + (off2 / resolution)).rgb * 0.0702702703;
    color += texture2D(image, uv - (off2 / resolution)).rgb * 0.0702702703;
    return color;
}

vec3 doBlur(float r, vec2 coord) {
    // vec3 size = vec3(viewWidth, viewHeight, r);

    // vec2 radius = size.z/size.xy;

    // for( float d=0.0;d<TWOPI;d+=TWOPI/float(DIRECTIONS) )
    // {
    //     for( float i=1.0/float(QUALITY);i<=1.0;i+=1.0/float(QUALITY) )
    //     {
    //         vec2 pos = coord+vec2(cos(d),sin(d))*radius*i;
    //         vec3 original = texture2D( gcolor, pos).rgb;
    //         original *= 0.6;
    //         color += doFog(original, pos);
    //     }
    // }
     //color /= float(QUALITY)*float(DIRECTIONS)+1.0;
    vec3 color = gauss9(gcolor, coord, r);
    return color;
}

vec3 doBloomBlur(float r, vec2 coord) {
    // vec3 size = vec3(viewWidth, viewHeight, r);

    // vec2 radius = size.z/size.xy;

    // for( float d=0.0;d<TWOPI;d+=TWOPI/float(BLOOM_DIRECTIONS) )
    // {
    //     for( float i=1.0/float(BLOOM_QUALITY);i<=1.0;i+=1.0/float(BLOOM_QUALITY) )
    //     {
    //         vec2 pos = coord+vec2(cos(d),sin(d))*radius*i;
    //         vec3 original = texture2D(colortex1, pos).rgb;

    //         color += original;
    //     }
    // }
    // color /= float(BLOOM_QUALITY)*float(BLOOM_DIRECTIONS)+1.0;
    vec3 color = vec3(0);
    float iterations = 18.0;
    for(int i = 0; i < iterations; i++) {
        float radius = (iterations - i - 1.0) * r * 0.07;
        vec2 dir;
        if(mod(i, 2.0) == 0.0) {
            dir = vec2(radius, 0);
        } else {
            dir = vec2(0, radius);
        }
        color += blur9(colortex1, coord, dir);
    }
    color /= iterations;
    return color;
}

vec2 onScreen(vec3 positionCameraSpace) {
    vec4 positionNdcSpace = gbufferProjection * vec4(positionCameraSpace, 1.0);
    positionNdcSpace.xyz /= positionNdcSpace.w;
    
    return positionNdcSpace.xy * 0.5 + 0.5;
}

vec4 sampleSun(vec2 pos) {
    if(texture2D(depthtex1, pos).r == 1.0) {
        vec2 screen = vec2(viewWidth, viewHeight);

        vec2 sunPos = onScreen(sunPosition);
        vec2 moonPos = onScreen(moonPosition);

        float sunDist = distance(pos*screen,sunPos*screen)/viewWidth;
        float moonDist = distance(pos*screen,moonPos*screen)/viewWidth;

        //Scale sun size with FOV
        sunDist *= 1.0/gbufferProjection[1][1] * 2.0;
        moonDist *= 1.0/gbufferProjection[1][1] * 2.0;

        if(sunPosition.z < 0) {
            sunDist = (0.03 - sunDist) / 0.03;
            sunDist = clamp(sunDist, 0.0, 1.0);
            float intensity = sunDist * 5.0;

            return clamp(vec4(sunColor.rgb, sunColor.a*intensity), 0.0, 1.0);
        } else {
            moonDist = (0.04 - moonDist) / 0.04 / 2.5;
            moonDist = clamp(moonDist, 0.0, 1.0);
            float intensity = moonDist * 7.0;

            return clamp(vec4(moonColor.rgb, moonColor.a*intensity), 0.0, 1.0);
        }
    } else {
        return vec4(0.0);
    }
}

vec3 doGodRays(vec2 pos) {
    vec3 color = vec3(0.0);
    
    vec2 sunPos = onScreen(sunPosition);

    vec2 dirFromSun = pos - sunPos;
    float samples = float(GODRAY_SAMPLES);
    //samples = 4.0;

    float dither = texture2D(gaux1, pos * 4096).r;


    vec2 dtc = (pos - sunPos) * (1. / samples);
    for (int i = 0; i < samples; i++) {
        pos -= dtc;

        vec2 samplePos = pos + (dtc*dither);
        vec4 sun = sampleSun(samplePos);
        color += sun.rgb * sun.a;

    }

    return color / samples;
}

vec3 getSceneColor(vec2 coord) {
    vec3 color;
    //Apply blur
    float f = getFogAmount(getSceneDist(coord));

    if(LIGHT_SCATTERING_INTENSITY > 0.0) {
        color = doBlur(LIGHT_SCATTERING_INTENSITY * (1-f), coord);
    }
    /* POST */
    color *= 0.75;

    /* Apply SSAO */
    #ifdef SSAO
        color *= calcAO(coord);
    #endif
    /* Apply fog */
    color = doFog(color, coord);

    /* Add bloom'd lights */
    if(BLOOM_INTENSITY > 0.0) {
        vec3 lights = doBloomBlur(20.0, coord);

        color += lights * BLOOM_INTENSITY;
    }
    return color;
}

vec3 worldToView(vec3 worldPos, float isPosition) {
    vec4 pos;
    pos.xyz = worldPos;
    pos.w = isPosition;
    //pos.xyz = pos.xyz * 2.0 - 1.0; //convert from the 0-1 range to the -1 to +1 range
    pos = gbufferModelView * pos;

    return pos.xyz;
}

vec3 viewToWorld(vec3 viewPos, float isPosition) {
    vec4 pos;
    pos.xyz = viewPos;
    pos.w = isPosition;
    //pos.xyz = pos.xyz * 2.0 - 1.0; //convert from the 0-1 range to the -1 to +1 range
    pos = gbufferModelViewInverse * pos;

    return pos.xyz;
}

bool basicallyEqual(vec3 one, vec3 two) {
    float dx = abs(one.x - two.x);
    float dy = abs(one.y - two.y);
    float dz = abs(one.z - two.z);
    float thresh = 0.001;
    return (dx < thresh && dy < thresh && dz < thresh);
}

vec4 getSunReflection(vec2 coord, float specularity, float reflCoeff, float ang) {
    vec3 initialPoint = getPointInView(coord);
    vec3 incidentDir = normalize(initialPoint);
    vec3 norm = texture2D(colortex2, coord).rgb;
    norm = norm * 2.0 - 1.0;
    vec3 rayDir = reflect(incidentDir, norm);

    float cosTheta = 1.0 - dot(norm, rayDir);

    float fresnel = reflCoeff + (ang - reflCoeff) * cosTheta * cosTheta * cosTheta * cosTheta * cosTheta;

    float fogType;//0 = night, 1 = day

    if(worldTime > 22000 && worldTime < 23999) {
        //From night to day
        fogType = map(float(worldTime), 22000.0, 23999.0, 0.0, 1.0);
    } else if(worldTime > 12000 && worldTime < 14000) {
        //From day to night
        fogType = map(float(worldTime), 12000.0, 14000.0, 1.0, 0.0);
    } else if(worldTime >= 0 && worldTime <= 12000) {
        //Day
        fogType = 1.0;
    } else {
        fogType = 0.0;
    }

    vec3 dirToSun = normalize(sunPosition - initialPoint);
    float dotProdSun = dot(rayDir, dirToSun);
    

    vec3 dirToMoon = normalize(moonPosition - initialPoint);
    float dotProdMoon = dot(rayDir, dirToMoon);

    vec4 sunSpec = vec4(vec3(0.0), fresnel);
    vec4 moonSpec = sunSpec;
        //float reflAmount = map(dotProdSun, 0.8, 1.0, 0.1, 1.0) * (1.0-wetness);
    float reflAmount = pow(dotProdSun, 15.0) * (1.0-wetness);
    sunSpec = max(vec4(vec3(1)*reflAmount, fresnel), 0.0);

    reflAmount = pow(dotProdMoon, 15.0) * (1.0-wetness);
    moonSpec = max(vec4(vec3(0.6, 0.6, 0.7)*reflAmount, fresnel), 0.0);

//mix(moonSpec, sunSpec, fogType)

    return mix(moonSpec, sunSpec, fogType);
}


vec4 getReflectionColor(vec2 coord, float specularity, float reflCoeff, float ang, bool isWater) {
    vec3 up = worldToView(vec3(0.0, 1.0, 0.0), 0.0);

    vec3 initialPoint = getPointInView(coord);
    vec3 incidentDir = normalize(initialPoint);
    vec3 norm = texture2D(colortex2, coord).rgb;
    norm = norm * 2.0 - 1.0;

    vec3 normWorld = viewToWorld(norm, 0.0);

    //Make non upwards faces during rain not reflective
    if(!isWater && normWorld.y < 0.4 && wetness > minWetness) {
        return vec4(vec3(0.0), 0.0);
    }

    vec3 rayDir = reflect(incidentDir, norm);

    //Check if ray is reflecting into water
    float downFacing = dot(rayDir, -up);

    if(downFacing < 0.3 && downFacing > 0) {
        rayDir = viewToWorld(rayDir, 0.0);
        rayDir.y *= -1;
        rayDir = worldToView(rayDir, 0.0);
    }

    //ripples during rain
    if(!isWater) {
        vec2 random = texture2D(noisetex, initialPoint.xz + frameTimeCounter*sign(sin(frameTimeCounter*100))).rb;
        rayDir.xz += random * 0.05 * rainStrength;
    }

    vec3 rayPos = initialPoint;

    vec3 fogColor = getFogColor();

    float cosTheta = 1.0 - dot(norm, rayDir);

    float fresnel = reflCoeff + (ang - reflCoeff) * cosTheta * cosTheta * cosTheta * cosTheta * cosTheta;

    if(!isWater && fresnel > 0.8) {
        fresnel = 0.8;
    }



    //If point is too far or ray going towards camera, dont do specular
    if(length(initialPoint) > MAX_REFLECTION_DIST || rayDir.z > 0.0) {
        return vec4(fogColor, fresnel);
    }



    //March ray
    float dist = 0.2;

    #if SSR_QUALITY != 0

        int MAX_MARCHES = 28;
        float MARCH_DISTANCE = 3.0;

        #if SSR_QUALITY == 2
            MAX_MARCHES = 84;
            MARCH_DISTANCE = 0.8;
        #elif SSR_QUALITY == 3
            MAX_MARCHES = 150;
            MARCH_DISTANCE = 0.5;
        #endif

        float BLEND_AMOUNT;

        #if MAX_REFLECTION_DIST == 35
            BLEND_AMOUNT = 8.0;
        #elif MAX_REFLECTION_DIST == 65
            BLEND_AMOUNT = 16.0;
        #else
            BLEND_AMOUNT = 32.0;
        #endif

        if(!isWater) {
            MAX_MARCHES = int(float(MAX_MARCHES)*0.5);
        }
       
        for(int i = 0; i < MAX_MARCHES; i++) {
            

            rayPos = initialPoint + rayDir*dist;

            //Check if ray is colliding
            vec2 rayOnScreen = onScreen(rayPos);
            if(rayOnScreen.xy != clamp(rayOnScreen.xy, 0.0, 1.0)) {
                //ray went off screen
                return vec4(fogColor, fresnel);
            }

            float actualDist = getSceneDist(rayOnScreen);
            float rayDistFromEye = length(rayPos);


            float hitGroundDist = length(initialPoint.xz);
            float actualGroundDist = length(getPointInViewOpaque(rayOnScreen).xz);

            //Check if ray collided and not on object in front of initial point
            if(actualDist < rayDistFromEye && hitGroundDist < actualGroundDist) {
                //Prevent close-object artifact

                //Blend between max reflection distance to avoid abrupt change
                float blend = (MAX_REFLECTION_DIST - length(initialPoint))/BLEND_AMOUNT;

                blend = min(blend, 1.0);
                blend = max(blend, 0.0);

                return mix(vec4(getSceneColor(rayOnScreen), fresnel), vec4(fogColor, fresnel), 1-blend);
            }
            //if(i < MAX_MARCHES*0.8) {
                dist += MARCH_DISTANCE;
            //} else {
            //    dist += MARCH_DISTANCE*18.0;
            //}
            
        }
    #endif
    return vec4(fogColor, fresnel);
}

vec2 UnderwaterRefraction(in vec2 coord) {
    // Created by: Niemand
    if(isEyeInWater == 1) {
        vec2 refraction = vec2(sin(frameTimeCounter * 1.75 + texcoord.x * 50.0 + texcoord.y * 25.0), cos(frameTimeCounter * 2.5 + texcoord.y * 100.0 + texcoord.x * 25.0));
        return coord + refraction * 0.002;
    }
    return coord;
}

void lookup(inout vec3 color, sampler2D lut) {
    const vec2 inverseSize = vec2(1.0 / 512, 1.0 / 512.0);
    const mat2 correctGrid = mat2(
        vec2(1.0, inverseSize.y * 512),
        vec2(0.0, inverseSize.y * 512.0)
    );

    // if (gl_FragCoord.x < 512.0 && gl_FragCoord.y < 512.0) {
    //     vec2 uv = gl_FragCoord.xy / 512.0;
    //     color = texture2D(gaux2, uv + correctGrid[1]).rgb;
    //     return;
    // }

    color = clamp(color, 0.0, 1.0);

    float blueColor = color.b * 63.0;

    vec4 quad = vec4(0.0);
    quad.y = floor(floor(blueColor) * 0.125);
    quad.x = floor(blueColor) - (quad.y * 8.0);
    quad.w = floor(ceil(blueColor) * 0.125);
    quad.z = ceil(blueColor) - (quad.w * 8.0);

    vec4 texPos = (quad * 0.125) + (0.123046875 * color.rg).xyxy + 0.0009765625;

    vec3 newColor1 = texture2D(lut, texPos.xy * correctGrid[0] + correctGrid[1]).rgb;
    vec3 newColor2 = texture2D(lut, texPos.zw * correctGrid[0] + correctGrid[1]).rgb;
    
    color = mix(newColor1, newColor2, fract(blueColor));
}

void ditherScreen(inout vec3 color) {
    vec3 lestynRGB = vec3(dot(vec2(171.0, 231.0), gl_FragCoord.xy));
         lestynRGB = fract(lestynRGB.rgb / vec3(103.0, 71.0, 97.0));

    color += lestynRGB.rgb / 255.0;
}

void main() {
    vec2 coord = texcoord.st;

    vec4 originalColour = texture2D(gcolor, coord);
    vec4 color = vec4(getSceneColor(coord), 1.0);

    
    float specularity;
    if(isEyeInWater > 0) {
        #ifdef UNDERWATER_REFRACTION
            coord = UnderwaterRefraction(coord);
        #endif
        specularity = 0.0;
    } else {
        specularity = texture2D(colortex3, coord).r;
    }

    float originalSpecularity = specularity;

    //Check if specular
    vec4 reflectionColor = vec4(0.0);
    if(originalSpecularity > 0.0) {
        if(originalSpecularity > 0.999) {//Water
            reflectionColor = getReflectionColor(coord, specularity, waterReflCoeff, 1.0, true);
            //reflectionColor.rgb = doFogForSpecular(reflectionColor.rgb, coord);
            specularity *= reflectionColor.w;
            color = mix(color, reflectionColor, specularity);
        } else if(wetness > minWetness){
            reflectionColor = getReflectionColor(coord, specularity, 0.11, 3.2, false);
            specularity *= reflectionColor.w;

            color = mix(color, reflectionColor, specularity);
        }

        
    }

    //Sun reflections
    vec4 sunRefl;
    if(originalSpecularity < 0.999) {//Water
        
        sunRefl = getSunReflection(coord, specularity, 0.81, 0.81);
    }
    
    if(reflectionColor.w < 0.00001) {
        //If ssr is not done, then apply fresnel from sun refl onto specularity
        specularity *= sunRefl.w;
    }
    float f = getFogAmount(getSceneDist(coord));

    color.rgb += sunRefl.rgb*specularity*f;
    //color.rgb += vec3(0.4) * specularity;

    vec4 sun = sampleSun(coord);

    #ifdef GODRAYS
        color.rgb += doGodRays(coord)*0.2;
    #else
        color.rgb += sun.rgb * sun.a *0.1;
    #endif
    vec4 handColour = texture2D(colortex4, coord);

    if(length(handColour) > 0.0) {
        //color = originalColour*0.8;
    }

    //color.rgb = texture2D(colortex4, coord).rgb;
    //color.rgb = sampleSun(coord);
    //color.rgb = vec3(bayer8192(coord*512));
    //color.rgb = vec3(calcAO(coord));
    #if LUT == 1
        vec3 lookedUpColor = color.rgb;
        lookup(lookedUpColor, gaux2);
        color.rgb = mix(color.rgb, lookedUpColor, 1-f*0.7);
        ditherScreen(color.rgb);
    #elif LUT == 2
        vec3 lookedUpColor = color.rgb;
        lookup(lookedUpColor, gaux3);
        color.rgb = mix(color.rgb, lookedUpColor, 1-f*0.7);//
        ditherScreen(color.rgb);
    #endif
    gl_FragColor = color;

}