#version 120
uniform float viewWidth;
uniform float viewHeight;
uniform sampler2D lightmap;
uniform sampler2D texture;
uniform sampler2D noisetex;
uniform float wetness;         
uniform bool rainBiome;
uniform bool coldBiome;
uniform vec3 cameraPosition;
uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;

uniform sampler2D normals;
uniform sampler2D specular;

varying float entity;
varying vec2 lmcoord;
varying vec2 texcoord;
varying vec4 glcolor;
varying float isLightSource;

varying mat3 tbnMatrixView;
varying mat3 tbnMatrixWorld;

float map(float n, float start1, float stop1, float start2, float stop2) {
    return ((n - start1) / (stop1 - start1)) * (stop2 - start2) + start2;
}

float calculateSpecularMaps(vec2 coord) {
    float specular = texture2D(specular, coord).r;
    return specular;
}

vec3 getTangentNormals(vec2 coord) {
    vec4 normal = texture2D(normals,  coord) * 2.0 - 1.0;
    return vec3(normal.x, normal.y, normal.z);
}

vec3 worldToView(vec3 worldPos, float isPosition) {
    vec4 pos;
    pos.xyz = worldPos;
    pos.w = isPosition;
    //pos.xyz = pos.xyz * 2.0 - 1.0; //convert from the 0-1 range to the -1 to +1 range
    pos = gbufferModelView * pos;

    return pos.xyz;
}

void main() {
	vec4 color = texture2D(texture, texcoord) * glcolor;
	color *= texture2D(lightmap, lmcoord);

    //Make things darker during rain
    color = mix(color, color*0.75, wetness);

    vec4 justLightSources;

    if(isLightSource > 0.0) {
        justLightSources = color;
    } else {
        justLightSources = vec4(0.0);
    }

    vec3 eye = cameraPosition + gbufferModelViewInverse[3].xyz;

    bool isRainyBiome = rainBiome || (coldBiome && eye.y < 90);

    float covered = 28.0/32.0;
    float uncovered = 31.0/32.0;

    float coverAmount = map(lmcoord.y, covered, uncovered, 0.0, 1.0);
    coverAmount = min(coverAmount, 1.0);
    coverAmount = max(coverAmount, 0.0);

    //Normal mapping
    vec3 normalTex = getTangentNormals(texcoord.st);
    vec3 normal = normalize(normalTex * tbnMatrixWorld);

    normal = worldToView(normal, 0.0);

    //Specular mapping
    float specularity = calculateSpecularMaps(texcoord.st);

    /* DRAWBUFFERS:0123 */
	gl_FragData[0] = color;//vec4(vec3(lmcoord.x), color.a); //gcolor
    gl_FragData[1] = justLightSources;
    gl_FragData[2] = vec4(normal, 1.0)* 0.5 + 0.5; // normals
if(wetness > 0.1) {
        gl_FragData[3] = vec4(0.6) * wetness * float(isRainyBiome) * coverAmount; // specularity when raining and exposed to rain
    } else {
        gl_FragData[3] = vec4(specularity); // specularity when not raining
    }
}