#version 120

#define WAVING_PLANTS //Plants and tree leaves swaying in the wind

uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjection;
uniform float frameTimeCounter; 
uniform float rainStrength;

varying vec2 lmcoord;
varying vec2 texcoord;
varying vec4 glcolor;
varying float isLightSource;
varying vec3 normal;
varying float entity;
varying mat3 tbnMatrixWorld;

attribute vec3 mc_Entity;
attribute vec4 mc_midTexCoord;
attribute vec4 at_tangent;


const float PI48 = 150.796447372;
float pi2wt = PI48*frameTimeCounter;

vec3 calcWaveLeaves(vec3 pos) {
    float magnitude = abs(sin(dot(vec4(frameTimeCounter, pos),vec4(1.0,0.005,0.005,0.005)))*0.5+0.72)*0.013;
    vec3 offset = (sin(pi2wt*vec3(0.0063,0.0224,0.0015)*1.5 - pos))*magnitude*1.6;

    offset.xz *= 2;

    return pos+offset;
}

vec3 calcWavePlants(in vec3 pos) {
    float magnitude = abs(sin(dot(vec4(frameTimeCounter, pos),vec4(1.0,0.005,0.005,0.005)))*0.5+0.72)*0.013;
    vec3 offset = (sin(pi2wt*vec3(0.0063,0.0224,0.0015)*1.5 - pos))*magnitude*1.6;

    offset.xz *= 4;
    offset.y *= 0.6;

    return pos + offset;
}

mat3 calculateTBN() {
	vec3 normal   = normalize(gl_NormalMatrix * gl_Normal);
	vec3 tangent  = normalize(gl_NormalMatrix * (at_tangent.xyz / at_tangent.w));

         normal   = mat3(gbufferModelViewInverse) * normal;
         tangent  = mat3(gbufferModelViewInverse) * tangent;

	vec3 binormal = normalize(cross(tangent, normal));

         return transpose(mat3(tangent, binormal, normal));
}

void main() {
    vec4 position = gbufferModelViewInverse * gl_ModelViewMatrix * gl_Vertex;

    #ifdef WAVING_PLANTS
        if(mc_Entity.x == 10005) {
            position.xyz = calcWavePlants(position.xyz);
        } else if(mc_Entity.x == 10002 && gl_MultiTexCoord0.t < mc_midTexCoord.t) {
            //Grass and flowers, only for vertices on top
            position.xyz = calcWavePlants(position.xyz);
        } else if(mc_Entity.x == 10003) {
            position.xyz = calcWavePlants(position.xyz);        
        }
    #endif
    gl_Position = gl_ProjectionMatrix * gbufferModelView * position;

	texcoord = (gl_TextureMatrix[0] * gl_MultiTexCoord0).xy;
	lmcoord  = (gl_TextureMatrix[1] * gl_MultiTexCoord1).xy;
	glcolor = gl_Color;

    if(mc_Entity.x == 10001) {
        isLightSource = 1.0;
    } else {
        isLightSource = 0.0;
    }

   // normal = normalize(gl_NormalMatrix * gl_Normal);
    // vec4 normalTex = getTangentNormals(texcoord.st);

    tbnMatrixWorld = calculateTBN();

    // normal  = normalize((normalTex.xyz) * tbnMatrixWorld);

    // normal = worldToView(normal, 0.0);

    entity = mc_Entity.x;
}