#version 120

uniform sampler2D noisetex;
uniform vec3 cameraPosition;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;
uniform sampler2D depthtex0;
uniform float frameTimeCounter; 

varying vec2 lmcoord;
varying vec2 texcoord;
varying vec4 glcolor;
varying vec3 normal;
varying vec3 noise;
varying float isWater;

attribute vec3 mc_Entity;

vec3 getNoise(vec2 pos) {
	vec2 time = vec2(sin(frameTimeCounter * 0.6), cos(frameTimeCounter * 0.6));

	vec2 offset0 = time*0.01;

	vec3 noise = texture2D(noisetex, pos / 128.0 + offset0).rgb * 0.5;

	vec2 offset1 = vec2(time.x*0.04, time.y*0.005);

	noise += texture2D(noisetex, pos / vec2(4.0, 256.0) - offset1).rgb * 0.8;

	vec2 offset2 = time*0.03;

	noise += texture2D(noisetex, pos / 1.5 + offset2).rgb * 0.5;

	return noise;
}

void main() {
	gl_Position = ftransform();
	texcoord = (gl_TextureMatrix[0] * gl_MultiTexCoord0).xy;
	lmcoord  = (gl_TextureMatrix[1] * gl_MultiTexCoord1).xy;
	glcolor = gl_Color;

	vec4 pos = gbufferModelViewInverse * (gl_ModelViewMatrix * gl_Vertex) + vec4(cameraPosition, 0.0);

	noise = getNoise(pos.xz) - 0.5;//= 0.5
	noise.y = 0;

	//Remove waving from stainedglass
	if(mc_Entity.x != 10004) {
		isWater = 0;
		normal = gl_Normal;
	} else {
		isWater = 1;
		normal = gl_Normal + noise*0.1;
	}

	

	normal = normalize(gl_NormalMatrix * normal);
}